/*
 * The Pipelined Regular Expression Monitor Compiler System
 * Copyright (c) 2002 University of British Columbia
 * All Rights Reserved.
 *
 *
 * Compiler for the PREMSC tool. Each function performs a reduction from
 * the grammar defined in grammar.y 
 *
 * Marcio Oliveira -- 2002
 */

#include <stdlib.h>
#include <error.h>
#include <math.h>
#include <compile.h>
#include <dd.h>
#include <translation.h>

extern int controlPointCount;

char *strdup();
float log2();

int defProd;
int numStateMachine;
stateMachinePtr stateMachineList;
char *topProduction;

DdManager *ddm;

void initCompile() {

    defProd = PDEF;
    topProduction = NULL;
    stateMachineList = NULL;
    numStateMachine = 0;

    ddm = bddInit();

    initSymbols();
}

signalPtr reduceSignal(char *n, int i0, int i1, int v, int l) {

    signalPtr s;
    int range;
    int i;
    int c;

    s = (signalPtr) malloc(sizeof(struct signal));
    memset(s, 0, sizeof(struct signal));
    s->name = strdup(n);
    s->index[0] = i0;
    s->index[1] = i1;
    s->initValue = v;
    s->srcline = l;
    if ((i0 != 0) || (i1 != 0)) {
        s->isArray = 1;
        c = ceil(log2((float)(v+1)));
        if ((abs(i0 - i1) + 1) < c)
            fatal_error("Invalid signal initialization."); 
    }
    range = abs(i0 - i1) + 1;
    if ((s->dd = (DdNode **) malloc(sizeof(DdNode *) * range)) == NULL)
        fatal_error("reduceSignal()");
    for (i=0; i<range; i++) {
        s->dd[i] = bddNewVar(ddm);
    }

    return s;
}

signalPtr reduceSignalList(signalPtr oldlist, signalPtr s) {

    if (s == NULL) fatal_error("Invalid declaration.");
    s->next = oldlist;

    return s;
}

void reduceSignalDeclarationList(signalType t, signalPtr list) {

    while (list != NULL) {
        list->type = t;
        switch (t) {
            case P_SIGNAL_INTERNAL:
                insertInternal(list);
                break;
            case P_SIGNAL_INPUT:
            case P_SIGNAL_OUTPUT:
            case P_SIGNAL_INOUT:
                insertSignal(list);
                break;
            default:
                fatal_error("reduceSignalDeclarationList()");
                break; 
        }
        list = list->next;
    } 
}

void reduceDefine(char *name, treeNodePtr expr, int srcline) { 
    
    definePtr d;
   
    d = (definePtr) malloc(sizeof(struct define));
    memset(d, 0, sizeof(struct define));
    d->name = strdup(name);
    d->dd = expr->dd;
    d->expr = expr;
    d->srcline = srcline;

    insertDefine(d);    
}

void reduceProduction(char *name, treeNodePtr prod, int srcline) {

    productionPtr p;

    p = (productionPtr) malloc(sizeof(struct production));
    memset(p, 0, sizeof(struct production));
    p->name = strdup(name);
    p->prod = prod;
    p->srcline = srcline;
    
    if (topProduction == NULL) {
        topProduction = strdup(name);
    }

    insertProduction(p);
}

void reduceStateMachine(char *init) {

    stateMachinePtr sm;

    sm = (stateMachinePtr) malloc(sizeof(struct stateMachine));
    memset(sm, 0, sizeof(struct stateMachine));
    sm->topProduction = strdup(init);
    sm->next = stateMachineList;
    sm->index = numStateMachine++;
    stateMachineList = sm;
}

actionPtr reduceAction(char *name, indexNodePtr index, actionNodePtr expr) {

    actionPtr a;

    a = (actionPtr) malloc(sizeof(struct action));
    if (a == NULL)
        fatal_error("reduceAction()");
 
    memset(a, 0, sizeof(struct action));

    if (index == NULL) {
        a->lhs = reduceActionIdent(name);
    } else {
        a->lhs = reduceActionIdentArray(name, index);
    }

    a->rhs = expr;

    return a;
}

actionPtr reduceActionList(actionPtr oldlist, actionPtr action) {

    actionPtr a;

    if (action == NULL) 
        fatal_error("Invalid action.");
  
    if (oldlist == NULL)
        return action;

    a = oldlist;
    while(a->next != NULL) {
        a = a->next;
    }

    a->next = action;
    return a;
}

actionNodePtr reduceActionIdent(char *name) {

    actionNodePtr an;
    signalPtr s;

    if (lookupInternal(name, &s) == PNOTFOUND)
        if (lookupSignal(name, &s) == PNOTFOUND)
            fatal_error("Undeclared var: %s\n.", name);

    an = (actionNodePtr) malloc(sizeof(struct actionNode));
    if (an == NULL)
        fatal_error("reduceActionIdent()");

    an->type = P_AN_SIGNAL;
    an->data.signal = s;

    return an;
}

actionNodePtr reduceActionIdentArray(char *name, indexNodePtr an1) {

    actionNodePtr an;
    signalPtr s;

    if (lookupInternal(name, &s) == PNOTFOUND)
        if (lookupSignal(name, &s) == PNOTFOUND)
            fatal_error("Undeclared var: %s\n.", name);

    if (!s->isArray)
        fatal_error("Invalid expression, %s is not an array.", name);

    an = (actionNodePtr) malloc(sizeof(struct actionNode));
    if (an == NULL)
        fatal_error("reduceActionIdentArray()");

    an->type = P_AN_SIGNAL_ARRAY;
    an->data.signal = s;
    an->index = an1;

    return an;
}

actionNodePtr reduceActionConstant(int c) {

    actionNodePtr an;

    an = (actionNodePtr) malloc(sizeof(struct actionNode));
    if (an == NULL)
        fatal_error("reduceActionConstant()");

    memset(an, 0, sizeof(struct actionNode));

    an->type = P_AN_CONSTANT;
    an->data.constant = c;

    return an; 
}

actionNodePtr reduceActionMinus(actionNodePtr an1, actionNodePtr an2) {

    actionNodePtr an;

    an = (actionNodePtr) malloc(sizeof(struct actionNode));
    if (an == NULL)
        fatal_error("reduceActionMinus()");

    memset(an, 0, sizeof(struct actionNode));

    an->type = P_AN_MINUS;
    an->left = an1;
    an->right = an2;

    return an;
}

actionNodePtr reduceActionPlus(actionNodePtr an1, actionNodePtr an2) {

    actionNodePtr an;

    an = (actionNodePtr) malloc(sizeof(struct actionNode));
    if (an == NULL)
        fatal_error("reduceActionPlus()");

    memset(an, 0, sizeof(struct actionNode));

    an->type = P_AN_PLUS;
    an->left = an1;
    an->right = an2;

    return an;
}

treeNodePtr reduceActionp(treeNodePtr p, actionPtr actionList) {
        
    if (p->type == P_TN_AT_P) 
        fatal_error("Actions can not be associated with @ operator.");
        
    p->actionList = actionList;
        
    return p; 
}

indexNodePtr reduceIndexIdent(char *name) {

    indexNodePtr in;
    signalPtr s;

    if (lookupInternal(name, &s) == PNOTFOUND)
        if (lookupSignal(name, &s) == PNOTFOUND)
            fatal_error("Undeclared var: %s\n.", name);

    in = (indexNodePtr) malloc(sizeof(struct indexNode));
    if (in == NULL)
        fatal_error("reduceIndexIdent()");

    in->type = P_IN_SIGNAL;
    in->data.signal = s;

    return in;
}

indexNodePtr reduceIndexConst(int value) {

    indexNodePtr in;

    in = (indexNodePtr) malloc(sizeof(struct indexNode));
    if (in == NULL)
        fatal_error("reduceIndexIdent()");

    in->type = P_IN_CONSTANT;
    in->data.constant = value;

    return in;
}

treeNodePtr reduceVarProd(char *name) {

    signalPtr s = NULL;
    definePtr d = NULL;
    treeNodePtr t = NULL;

    switch(defProd) {
        case PDEF:
               if (lookupSignal(name, &s) == POK) {
                   if (s->isArray)
                       fatal_error(
                      "Signal arrays can only be used in comparison \"==\".\n");
                   t = createTreeNode(s, s->dd[0], P_TN_SIGNAL, NULL, NULL);
               } else if (lookupInternal(name, &s) == POK) {
                   if (s->isArray)
                       fatal_error(
                      "Signal arrays can only be used in comparison \"==\".\n");
                   t = createTreeNode(s, s->dd[0], P_TN_INTERNAL, NULL, NULL);
               } else if (lookupDefine(name, &d) == POK) {
                   t = createTreeNode(d, d->dd, P_TN_DEFINE, NULL, NULL);
               } else {
                   fatal_error("Undeclared signal or define: %s\n", name);
               }   
            break;
        case PPROD:
               if (lookupSignal(name, &s) == POK) {
                   if (s->isArray)
                       fatal_error(
                      "Signal arrays can only be used in comparison \"==\".\n");
                   t = createTreeNode(s, s->dd[0], P_TN_SIGNAL, NULL, NULL);
               } else if (lookupInternal(name, &s) == POK) {
                   if (s->isArray)
                       fatal_error(
                      "Signal arrays can only be used in comparison \"==\".\n");
                   t = createTreeNode(s, s->dd[0], P_TN_INTERNAL, NULL, NULL);
               } else if (lookupDefine(name, &d) == POK) {
                   t = createTreeNode(d, d->dd, P_TN_DEFINE, NULL, NULL);
               } else {
                   t = createTreeNode(strdup(name), NULL, P_TN_PROD, 
                                                                    NULL, NULL);
               }
            break;
        default:
            fatal_error("reduceVarProd()\n");
            break;
    }

    return t;
}


treeNodePtr reduceVar(char *name) {

    signalPtr s = NULL;
    treeNodePtr t = NULL;

    if (lookupSignal(name, &s) == POK) {
        t = createTreeNode(s, s->dd[0], P_TN_SIGNAL, NULL, NULL);
    } else if (lookupInternal(name, &s) == POK) {
        t = createTreeNode(s, s->dd[0], P_TN_INTERNAL, NULL, NULL);
    } else {
        fatal_error("Undeclared signal: %s\n", name);
    }   

    return t;
}

treeNodePtr reduceVarArray(char *name, indexNodePtr index) {

    signalPtr s, n;
    treeNodePtr t;

    if (lookupSignal(name, &s) == POK) {
        n = (signalPtr) malloc(sizeof(struct signal));
        memcpy(n, s, sizeof(struct signal));
        n->indexN = index;

        t = createTreeNode(n, (void *) 1, P_TN_SIGNAL_ARRAY, NULL, NULL);

    } else if (lookupInternal(name, &s) == POK) {
        n = (signalPtr) malloc(sizeof(struct signal));
        memcpy(n, s, sizeof(struct signal));
        n->indexN = index;

        t = createTreeNode(n, (void *) 1, P_TN_INTERNAL_ARRAY, NULL, NULL);

    } else { 
        fatal_error("Undeclared signal: %s\n", name);
    }

    return t;
}

treeNodePtr reduceConstant(int c) {

    return createTreeNode((void *)c, (void *) 1, P_TN_CONSTANT, NULL, NULL);
}


treeNodePtr reduceEqualEqual(treeNodePtr c1, treeNodePtr c2) {

    treeNodePtr t;

    checkComparison(c1, c2);

    t = createTreeNode(NULL, (void *) 1, P_TN_EQUALEQUAL, c1, c2);

    return t;
}

treeNodePtr reduceNotEqual(treeNodePtr c1, treeNodePtr c2) {

    treeNodePtr t;

    checkComparison(c1, c2);

    t = createTreeNode(NULL, (void *) 1, P_TN_NOTEQUAL, c1, c2);

    return t;
}

treeNodePtr reduceAndb(treeNodePtr expr1, treeNodePtr expr2) {

    treeNodePtr t;

    if (expr1->type == P_TN_PROD)
        fatal_error("Production %s can not be used in boolean expressions.\n",
                    (char *) expr1->data);

    if (expr2->type == P_TN_PROD)
        fatal_error("Production %s can not be used in boolean expressions.\n",
                    (char *) expr2->data);
    
    t = createTreeNode(NULL, (void *) 1, P_TN_AND_B, expr1, expr2);

/*
    t = createTreeNode(NULL, bddAnd(ddm, expr1->dd, expr2->dd), P_TN_AND_B,
                                                                 expr1, expr2);
*/

    return t;
}

treeNodePtr reduceOrb(treeNodePtr expr1, treeNodePtr expr2) {

    treeNodePtr t;

    if (expr1->type == P_TN_PROD)
        fatal_error("Production %s can not be used in boolean expressions.\n",
                    (char *) expr1->data);

    if (expr2->type == P_TN_PROD)
        fatal_error("Production %s can not be used in boolean expressions.\n",
                    (char *) expr2->data);

    t = createTreeNode(NULL, (void *) 1, P_TN_OR_B, expr1, expr2); 

/*
    t = createTreeNode(NULL, bddOr(ddm, expr1->dd, expr2->dd), P_TN_OR_B,
                                                                  expr1, expr2);
*/

    return t;
}

treeNodePtr reduceNotb(treeNodePtr expr) {

    treeNodePtr t;

    if (expr->type == P_TN_PROD)
        fatal_error("Production %s can not be used in boolean expressions.\n",
                    (char *) expr->data);

    t = createTreeNode(NULL, (void *) 1, P_TN_NOT_B, expr, NULL); 

/*
    t = createTreeNode(NULL, bddNot(expr->dd), P_TN_NOT_B, expr, NULL);
*/

    return t;
}

treeNodePtr reduceOrp(treeNodePtr p1, treeNodePtr p2) {
       
    treeNodePtr t;
           
    t = createTreeNode(NULL, NULL, P_TN_OR_P, p1, p2);

    return t;
}

treeNodePtr reduceCommap(treeNodePtr p1, treeNodePtr p2) {

    treeNodePtr t;

    t = createTreeNode(NULL, NULL, P_TN_COMMA_P, p1, p2);

    return t;
}

treeNodePtr reduceHatp(treeNodePtr p, int n) {

    treeNodePtr t;
    int i;

    if (n <= 0)
        fatal_error("Invalid index for operator \"^\".\n");

    if (n == 1)
        return p;

    t = p;
    for (i=1; i<n; i++) {
        t = reduceCommap(t, p);
    }

    return t;
}

treeNodePtr reduceHatpInterval(treeNodePtr p, int n1, int n2) {

    treeNodePtr t;
    treeNodePtr *tempArray;
    int rep;
    int i;

    if ((n1 <= 0) || (n2 <= 0))
        fatal_error("Invalid index for operator \"^\".\n");

    if (n1 > n2)
        fatal_error("Invalid index range for operator \"^\".\n");

    if (n1 == n2)
        return reduceHatp(p, n1);

    rep = (n2 - n1) + 1;

    tempArray = (treeNodePtr *) malloc(sizeof(treeNodePtr)*rep);
    memset(tempArray, 0, sizeof(treeNodePtr)*rep);

    for(i=0; i<rep; i++) {
        tempArray[i] = reduceHatp(p, n1+i);
    }

    for (i= 1, t = tempArray[0]; i < rep; i++) {
        t = reduceOrp(t, tempArray[i]);
    }

    free(tempArray);

    return t;
}

treeNodePtr reduceStarp(treeNodePtr p) {

    treeNodePtr t;

    t = createTreeNode(NULL, NULL, P_TN_STAR_P, p, NULL);

    return t;
}

treeNodePtr reduceAtp(treeNodePtr p1, treeNodePtr p2) {

    treeNodePtr t;

    t = createTreeNode(NULL, NULL, P_TN_AT_P, p1, p2);

    return t;
}

void createParseTrees() {

    stateMachinePtr sm;
    hashTablePtr recursionTable;
    productionPtr p;
    treeNodePtr ok;
    treeNodePtr okp;
    treeNodePtr en;
    treeNodePtr temp;
    treeNodePtr false;
    treeNodePtr okfc, acfc, enfc;

    if (stateMachineList == NULL) {
       stateMachineList = (stateMachinePtr) malloc(sizeof(struct stateMachine));
       memset(stateMachineList, 0 , sizeof(struct stateMachine)); 
       stateMachineList->topProduction = topProduction;
   }   

   sm = stateMachineList;
   while(sm != NULL) {
      if (lookupProduction(sm->topProduction, &p) == PNOTFOUND)
          fatal_error("Invalid production: %s.", sm->topProduction);

      recursionTable = createHashTable(RECURSION_HASH_SIZE);
      insertHashTableEntry(recursionTable, strHash, strCompare,
                           sm->topProduction, NULL);
 
      sm->parseTree = p->prod; 
      createParseTreesAux(&(sm->parseTree), recursionTable);
      sm->cp[0] = controlPointCount;
      temp = createTreeNode(createControlPoint(), NULL, P_TN_CP, NULL, NULL);
      false = createTreeNode(NULL, NULL, P_TN_FALSE_C, NULL, NULL);
      sm->recognizer = build(sm->parseTree, temp, temp, &ok, &okp, &en, &okfc, &acfc, &enfc);
      createExThread(ok, temp, en, false);
      sm->cp[1] = controlPointCount-1;
      sm = sm->next;
   } 
}

DdNode** createParseTreesAux(treeNodePtr *node, hashTablePtr rt) {

    productionPtr p;
    actionPtr a;
    DdNode **h[2];
    DdNode **result;

    if (isBoolExpr(*node)) {
        result = (DdNode **) malloc(sizeof(DdNode *) * 4);
        if (result == NULL)
            fatal_error("createParseTreesAux()");
 
        result[0] = (*node)->dd;
        result[1] = (*node)->dd;
        result[2] = bddZero(ddm);
        result[3] = bddZero(ddm);
        return result; 
    }

    if ((*node)->type == P_TN_PROD) {
        if (lookupProduction((char *)(*node)->data, &p) == PNOTFOUND)
            fatal_error("Invalid production: %s.", (char *)(*node)->data);

        if (insertHashTableEntry(rt, strHash, strCompare, p->name, NULL) == 
                                                                PDUPLICATED_KEY)
            fatal_error("Recursion of productions is not allowed.");

        if ((*node)->actionList != NULL) {
            if (p->prod->actionList != NULL) {
                a = (*node)->actionList;
                while (a->next != NULL) {
                    a = a->next;
                } 
                a->next = p->prod->actionList; 
            }
            p->prod->actionList = (*node)->actionList;
        } 

        *node = p->prod; 
        result = createParseTreesAux(node, rt);

        removeHashTableEntry(rt, strHash, strCompare, p->name);
    } else {
        if ((*node)->left != NULL)
           h[0] = createParseTreesAux(&((*node)->left), rt);
        if ((*node)->right != NULL)
           h[1] = createParseTreesAux(&((*node)->right), rt);
/*
        result = checkNonDeterminism(*node, h);
*/
        result = NULL;
    }

    return result;
}

DdNode** checkNonDeterminism(treeNodePtr node, DdNode ***h) {

    DdNode **result;
    DdNode *temp;

    result = (DdNode **) malloc(sizeof(DdNode *) * 4);
    if (result == NULL)
        fatal_error("checkNonDeterminism()");

    switch (node->type) {
        case P_TN_OR_P:
            temp = bddAnd(ddm, h[0][0], h[1][0]);
            if (temp != bddZero(ddm)) {
                fatal_error("Invalid non-determnistic OR.");
            } 
            bddFree(ddm, temp);

            result[0] = bddOr(ddm, h[0][0], h[1][0]);
            result[1] = bddOr(ddm, h[0][1], h[1][1]);
            result[2] = bddOr(ddm, h[0][2], h[1][2]);
            result[3] = bddOr(ddm, h[0][3], h[1][3]);
            break;
        case P_TN_COMMA_P:
            temp = bddAnd(ddm, h[0][3], h[1][0]);
            if (temp != bddZero(ddm)) {
                fatal_error("Invalid non-determnistic STAR COMMA.");
            } 
            bddFree(ddm, temp); 

            if ((node->left->type == P_TN_STAR_P) && 
                (node->right->type == P_TN_STAR_P)) {
                result[0] = bddOr(ddm, h[0][0], h[1][0]);
                result[3] = h[1][3];
            } else {
                result[0] = h[0][0];
                result[3] = bddZero(ddm); 
            }
            result[1] = h[1][1];
            result[2] = h[1][2];
            break;
        case P_TN_STAR_P:
            if (!isBoolExpr(node->left)) {
                temp = bddAnd(ddm, h[0][0], h[0][2]);
                if (temp != bddZero(ddm)) {
                    fatal_error("Invalid non-determnistic STAR STAR.");
                }
                bddFree(ddm, temp);
            }
            result[0] = h[0][0];
            result[1] = h[0][1];
            result[2] = h[0][1];
            result[3] = h[0][0]; 
            break;
        case P_TN_AT_P:
            temp = bddAnd(ddm, h[0][3], h[1][0]);
            if (temp != bddZero(ddm)) {
                fatal_error("Invalid non-determnistic STAR PIPELINE.");
            }
            bddFree(ddm, temp);

            result[0] = h[0][0];
            result[1] = h[0][1];
            result[2] = h[0][2];
            result[3] = h[0][3];
            break;
        default:
            fatal_error("checkNonDeterminism");
    }

    return result;      
}

int isBoolExpr(treeNodePtr node) {

    if (node->dd != NULL) 
        return 1;
    else 
        return 0; 
}

treeNodePtr createTreeNode(void *data, DdNode *dd, int type,
                           treeNodePtr left, treeNodePtr right) {

    treeNodePtr t;

    if ((t = (treeNodePtr) malloc(sizeof(struct treeNode))) == NULL) 
        fatal_error("createTreeNode()");

    memset(t, 0, sizeof(struct treeNode));
    t->data = data;
    t->dd = dd;
    t->type = type;
    t->left = left;
    t->right = right;

    return t;
}

void checkComparison(treeNodePtr tn1, treeNodePtr tn2) {

    signalPtr s1, s2;
    int c1, c2;
    int c1s, c2s;

    s1 = (signalPtr) tn1->data;
    s2 = (signalPtr) tn2->data;

    c1 = (int) tn1->data;
    c2 = (int) tn2->data;
    c1s = ceil(log2((float)(c1+1)));
    c2s = ceil(log2((float)(c2+1)));

    switch (tn1->type) {
        case P_TN_SIGNAL:
        case P_TN_INTERNAL:
            switch (tn2->type) {
                case P_TN_SIGNAL:
                case P_TN_INTERNAL:
                    if ((s1->index[0] != s2->index[0]) || 
                        (s1->index[1] != s2->index[1]))
                        fatal_error("Signal sizes don't match."); 
                    break;
                case P_TN_SIGNAL_ARRAY:
                case P_TN_INTERNAL_ARRAY:
                    if (s1->isArray)
                        fatal_error("Invalid comparison."); 
                    break;
                case P_TN_CONSTANT:
                    if ((abs(s1->index[0] - s1->index[1]) + 1) < c2s)
                        fatal_error("Signal size is smaller than size required by constant."); 
                    break;
                default:
                    fatal_error("checkComparison()"); 
            }
            break;

        case P_TN_SIGNAL_ARRAY:
        case P_TN_INTERNAL_ARRAY:
            switch (tn2->type) {
                case P_TN_SIGNAL:
                case P_TN_INTERNAL:
                    if (s2->isArray)
                        fatal_error("Invalid comparison.");
                    break;  
                case P_TN_SIGNAL_ARRAY:
                case P_TN_INTERNAL_ARRAY:
                    break;
                case P_TN_CONSTANT:
                    if (c2s > 1) 
                        fatal_error("Signal size is smaller than size required by constant.");
                    break;
                default:
                    fatal_error("checkComparison()");
            }
            break;

        case P_TN_CONSTANT:
            switch (tn2->type) {
                case P_TN_SIGNAL:
                case P_TN_INTERNAL:
                    if ((abs(s2->index[0] - s2->index[1]) + 1) < c1s)
                        fatal_error("Signal size is smaller than size required by constant.");
                    break;
                case P_TN_SIGNAL_ARRAY:
                case P_TN_INTERNAL_ARRAY:
                    if (c1s > 1)
                        fatal_error("Signal size is smaller than size required by constant.");
                    break;
                case P_TN_CONSTANT:
                    fatal_error("Invalid comparison expression.");
                    break;
                default:
                    fatal_error("checkComparison()");
            }
            break;

        default:
            fatal_error("checkComparison()");  
    }
}
