/*
 * The Pipelined Regular Expression Monitor Compiler System
 * Copyright (c) 2002 University of British Columbia
 * All Rights Reserved.
 *
 *
 * BDD functions
 *
 * Marcio Oliveira -- 2002
 */

#include <dd.h>
#include <error.h>

DdManager *bddInit() {
    DdManager *ddm;

    ddm = Cudd_Init(0,0,CUDD_UNIQUE_SLOTS,CUDD_CACHE_SLOTS,0);
    Cudd_AutodynEnable(ddm, CUDD_REORDER_SIFT); 

    return ddm;
}

DdNode *bddAnd(DdManager *ddm, DdNode *f, DdNode *g) {

    DdNode *r;

    r = Cudd_bddAnd(ddm, f, g);
    if (r == NULL)
        fatal_error("bddAnd()");

    Cudd_Ref(r);

    return r;
}

DdNode *bddOr(DdManager *ddm, DdNode *f, DdNode *g) {

    DdNode *r;

    r = Cudd_bddOr(ddm, f, g);
    if (r == NULL)
        fatal_error("bddOr()");

    Cudd_Ref(r);

    return r;
}

DdNode *bddNot(DdNode *f) {

    DdNode *r;

    r = Cudd_Not(f);
    if (r == NULL)
        fatal_error("bddNot()");

    Cudd_Ref(r);

    return r;
}

DdNode *bddOne(DdManager *ddm) {

    return Cudd_ReadOne(ddm);

}

DdNode *bddZero(DdManager *ddm) {

    return Cudd_ReadLogicZero(ddm);

} 

DdNode *bddNewVar(DdManager *ddm) {

    DdNode *r;
    static int i = 0;

    r = Cudd_bddNewVar(ddm);
    if (r == NULL) {
        printf("i: %d, ", i);
        i++;
        printf("i++: %d\n", i);
        fatal_error("bddNewVar()");
    }
    Cudd_Ref(r);

    i++;
    return r;
}

DdNode *bddCompose(DdManager *ddm, DdNode *f, DdNode *g, DdNode *v) {

    DdNode *r;

    r = Cudd_bddCompose(ddm, f, g, Cudd_NodeReadIndex(v));
    if (r == NULL) {
        fatal_error("bddCompose()");
    }
    Cudd_Ref(r);

    return r;
}

DdNode *bddElse(DdManager *ddm, DdNode *f) {

    DdNode *r;
    DdNode *temp;

    r = Cudd_E(f);
    if (r == NULL)
        fatal_error("bddElse()");
    Cudd_Ref(r);

    if (Cudd_IsComplement(f)) {
        temp = r;
        r = bddNot(temp);
        bddFree(ddm, temp);
    }

    return r;
} 

DdNode *bddThen(DdManager *ddm, DdNode *f) {

    DdNode *r;
    DdNode *temp;

    r = Cudd_T(f);
    if (r == NULL)
        fatal_error("bddThen()");
    Cudd_Ref(r);

    if (Cudd_IsComplement(f)) {
        temp = r;
        r = bddNot(temp);
        bddFree(ddm, temp);
    }

    return r;
}

DdNode *bddEquiv(DdManager *ddm, DdNode *f, DdNode *g) {

    DdNode *r;

    r = Cudd_bddXnor(ddm, f, g);

    if (r == NULL)
        fatal_error("bddEquiv()");

    Cudd_Ref(r);

    return r;
}

DdNode *bddEquivArray(DdManager *ddm, DdNode **f, DdNode **g, int size) {

    DdNode *r;
    DdNode *temp[2];
    int i;

    r = bddOne(ddm);

    for (i=0; i<size; i++) {
        temp[0] = bddEquiv(ddm, f[i], g[i]);
        temp[1] = bddAnd(ddm, r, temp[0]);
        bddFree(ddm, r);
        bddFree(ddm, temp[0]);
        r = temp[1];
    }

    return r;
}

void bddFree(DdManager *ddm, DdNode *f) {

    Cudd_RecursiveDeref(ddm, f);
}

