/*
 * The Pipelined Regular Expression Monitor Compiler System
 * Copyright (c) 2002 University of British Columbia
 * All Rights Reserved.
 *
 *
 * Management of the symbol tables for PREMSC 
 *
 * Marcio Oliveira -- 2002
 */

#include <hash.h>
#include <error.h>
#include <symbols.h>

hashTablePtr signalTable;
hashTablePtr internalTable;
hashTablePtr defineTable;
hashTablePtr productionTable; 

signalPtr signalList;
signalPtr internalList;
definePtr defineList;

signalPtr signalLast;
signalPtr internalLast;
definePtr defineLast;

void initSymbols() {

    signalTable = createHashTable(SIGNAL_HASH_SIZE);
    internalTable = createHashTable(INTERNAL_HASH_SIZE);
    defineTable = createHashTable(DEFINE_HASH_SIZE);
    productionTable = createHashTable(PRODUCTION_HASH_SIZE);

    signalList = NULL;
    internalList = NULL;
    defineList = NULL;
 
    signalLast = NULL;
    internalLast = NULL;
    defineLast = NULL;
}

void insertSignal(signalPtr s) {

    int result;
    signalPtr st;
    void *temp;

    result = insertHashTableEntry(signalTable, strHash, strCompare, s->name, s);
    if (result == PDUPLICATED_KEY)
        fatal_error("Duplicated declaration of signal %s at line %d.", 
                                                           s->name, s->srcline);
        
    result = lookupHashTableEntry(internalTable, strHash, strCompare, 
                                                                s->name, &temp);
    if (result == POK) {
        fatal_error("Duplicated declaration of signal %s at line %d.",
                                                           s->name, s->srcline);
    } 
    
    if ((st = (signalPtr) malloc(sizeof(struct signal))) == NULL)
        fatal_error("insertSignal()");
    memcpy(st, s, sizeof(struct signal));
    st->next = NULL;

    if (signalList == NULL) {
        signalList = st;
        signalLast = st;
    } else {
        signalLast->next = st;
        signalLast = st;  
    }
}

void insertInternal(signalPtr s) {

    int result;
    signalPtr st;
    void *temp;

    result = insertHashTableEntry(internalTable, strHash, 
                                  strCompare, s->name, s);
    if (result == PDUPLICATED_KEY)
        fatal_error("Duplicated declaration of signal %s at line %d.",
                                                           s->name, s->srcline);

    result = lookupHashTableEntry(signalTable, strHash, strCompare,
                                                                s->name, &temp);
    if (result == POK) {
        fatal_error("Duplicated declaration of signal %s at line %d.",
                                                           s->name, s->srcline);
    } 
    
    if ((st = (signalPtr) malloc(sizeof(struct signal))) == NULL)
        fatal_error("insertSignal()");
    memcpy(st, s, sizeof(struct signal));
    st->next = NULL;

    if (internalList == NULL) {
        internalList = st;
        internalLast = st;
    } else {
        internalLast->next = st;
        internalLast = st;
    }
}

void insertDefine(definePtr d) {

    int result;
    definePtr dt;
    void *temp;

    result = insertHashTableEntry(defineTable, strHash, strCompare, d->name, d);
    if (result == PDUPLICATED_KEY)
        fatal_error("Duplicated declaration of define %s at line %d.",
                                                           d->name, d->srcline);

    result = lookupHashTableEntry(signalTable, strHash, strCompare,
                                                                d->name, &temp);
    if (result == POK) {
        fatal_error("Duplicated declaration of define %s at line %d.",
                                                           d->name, d->srcline);
    }

    result = lookupHashTableEntry(internalTable, strHash, strCompare,
                                                                d->name, &temp);
    if (result == POK) {
        fatal_error("Duplicated declaration of define %s at line %d.",
                                                           d->name, d->srcline);
    }

    if ((dt = (definePtr) malloc(sizeof(struct define))) == NULL)
        fatal_error("insertSignal()");
    memcpy(dt, d, sizeof(struct define));
    dt->next = NULL;

    if (defineList == NULL) {
        defineList = dt;
        defineLast = dt;
    } else {
        defineLast->next = dt;
        defineLast = dt;
    }
}

void insertProduction(productionPtr p) {

    int result;
    void *temp; 

    result = insertHashTableEntry(productionTable, strHash, strCompare, 
                                                                    p->name, p);
    if (result == PDUPLICATED_KEY)
        fatal_error("Duplicated declaration of production %s at line %d.",
                                                           p->name, p->srcline);

    result = lookupHashTableEntry(signalTable, strHash, strCompare,
                                                                p->name, &temp);
    if (result == POK) {
        fatal_error("Duplicated declaration of production %s at line %d.",
                                                           p->name, p->srcline);
    }

    result = lookupHashTableEntry(internalTable, strHash, strCompare,
                                                                p->name, &temp);
    if (result == POK) {
        fatal_error("Duplicated declaration of production %s at line %d.",
                                                           p->name, p->srcline);
    }

    result = lookupHashTableEntry(defineTable, strHash, strCompare,
                                                                p->name, &temp);
    if (result == POK) {
        fatal_error("Duplicated declaration of production %s at line %d.",
                                                           p->name, p->srcline);
    }
}

int lookupSignal(char *name, signalPtr *s) {
   
    int result;
    
    result = lookupHashTableEntry(signalTable, strHash, strCompare,
                                                             name, (void **) s);
    return result;
}

int lookupInternal(char *name, signalPtr *s) {

    int result;

    result = lookupHashTableEntry(internalTable, strHash, strCompare,
                                                             name, (void **) s);
    return result;
}

int lookupDefine(char *name, definePtr *d) {

    int result;

    result = lookupHashTableEntry(defineTable, strHash, strCompare, 
                                                             name, (void **) d);
    return result;
}

int lookupProduction(char *name, productionPtr *p) {

    int result;

    result = lookupHashTableEntry(productionTable, strHash, strCompare, 
                                                             name, (void **) p);
    return result;
}

void checkIndex(signalPtr s, int index) {

    if ((index > s->index[0]) && (index > s->index[1]))
        fatal_error("Index out of range: %s[%d]\n", s->name, index);

    if ((index < s->index[0]) && (index < s->index[1]))
        fatal_error("Index out of range: %s[%d]\n", s->name, index);
}

int getOffset(signalPtr s) {

    if (s->index[0] < s->index[1])
        return s->index[0];
    else 
        return s->index[1];
}
