/*
 * The Pipelined Regular Expression Monitor Compiler System
 * Copyright (c) 2002 University of British Columbia
 * All Rights Reserved.
 *
 *
 * Verilog code generation.   
 *
 * Marcio Oliveira -- 2002
 */

#include <verilog.h>
#include <symbols.h>
#include <error.h>
#include <translation.h>

extern stateMachinePtr stateMachineList;

extern signalPtr signalList;
extern signalPtr internalList;
extern definePtr defineList;

extern controlPointPtr controlPointList;
extern int controlPointCount;

extern actionPointPtr actionPointList;
extern int actionPointCount;

extern exThreadPtr exThreadList;
extern int exThreadCount;

extern wirePtr wireList;
extern int wireCount;

void dumpVerilog(FILE *fd) {

    dumpVerilogModuleDeclaration(fd);
    dumpVerilogVarDeclaration(fd);
    dumpVerilogWire(fd);
    dumpVerilogAlways(fd);
    dumpVerilogEnd(fd);
}

void dumpVerilogModuleDeclaration(FILE *fd) {

    signalPtr s;

    fprintf(fd, "\nmodule MONITOR(");

    s = signalList;
    while (s != NULL) {
        fprintf(fd, "%s, ", s->name);
        s = s->next;
    }

    fprintf(fd, "clk, reset, ok);\n");

    s = signalList;
    while (s != NULL) {
        if (s->isArray) {
            fprintf(fd, "    input [%d:%d] %s;\n", s->index[0], 
                                                          s->index[1], s->name);
        } else {
            fprintf(fd, "    input %s;\n", s->name);
        }  
        s = s->next;
    }

    fprintf(fd, "    input clk;\n");
    fprintf(fd, "    input reset;\n");
    fprintf(fd, "    output ok;\n\n");

}


void dumpVerilogVarDeclaration(FILE *fd) {

    signalPtr s;

    s = signalList;
    while (s != NULL) {
        if (s->isArray) {
            fprintf(fd, "    wire [%d:%d] %s;\n", s->index[0],
                                                          s->index[1], s->name);
        } else {
            fprintf(fd, "    wire %s;\n", s->name);
        }
        s = s->next;
    }
    fprintf(fd, "    wire clk;\n");
    fprintf(fd, "    wire reset;\n");
    fprintf(fd, "    reg ok;\n");

    s = internalList;
    while (s != NULL) {
        if (s->isArray) {
            fprintf(fd, "    reg [%d:%d] %s;\n", s->index[0],
                                                          s->index[1], s->name);
        } else {
            fprintf(fd, "    reg %s;\n", s->name);
        }
        s = s->next;
    }

    fprintf(fd, "\n");

    fprintf(fd, "    reg [%d:0] pbs_active;\n", exThreadCount-1);
    fprintf(fd, "    reg [%d:0] pbs_enable;\n", exThreadCount-1);
    fprintf(fd, "    reg [%d:0] pbs_okn;\n", exThreadCount-1);
    fprintf(fd, "    reg [%d:0] pbs_reentry;\n", exThreadCount-1);
    fprintf(fd, "    reg [%d:0] pbs_ok;\n", exThreadCount-1);
    fprintf(fd, "    reg [%d:0] pbs_X;\n", controlPointCount-1);
    fprintf(fd, "    reg [%d:0] pbs_Y;\n", controlPointCount-1);
    fprintf(fd, "    wire [%d:0] pbs_w;\n", wireCount-1);
    if (actionPointCount != 0) 
        fprintf(fd, "    reg [%d:0] pbs_A;\n\n", actionPointCount-1);
}


void dumpVerilogWire(FILE *fd) {

    definePtr d;
    wirePtr w;
    
    d = defineList;
    while (d != NULL) {
        fprintf(fd, "    wire %s = ", d->name);
        dumpVerilogBoolExpression(fd, d->expr);
        fprintf(fd, ";\n");  
        d = d->next;
    }
    fprintf(fd, "\n");

    for (w = wireList; w != NULL; w = w->next) {
        fprintf(fd, "    assign pbs_w[%d] = ", w->index);
        dumpVerilogExpression(fd, w->f, "X");
        fprintf(fd, ";\n");
    }
    fprintf(fd, "\n");

}

void dumpVerilogAlways(FILE *fd) {

    controlPointPtr cp;
    actionPointPtr ap;
    exThreadPtr et;
    stateMachinePtr sm;
    signalPtr s;

    fprintf(fd, "    always @(posedge clk)\n");
    fprintf(fd, "    begin\n");

    fprintf(fd, "        if (reset == 1'b1)\n");
    fprintf(fd, "        begin\n");

    cp = controlPointList;
    while (cp != NULL) {
        fprintf(fd, "            pbs_X[%d] = 1'b", cp->index);
        if (cp->f == NULL)
            fprintf(fd, "1;\n");
        else 
            fprintf(fd, "0;\n"); 
        cp = cp->next;
    }

    fprintf(fd, "            ok = 1'b1;\n");

    for (s = internalList; s != NULL; s = s->next) {
        fprintf(fd, "            %s = %d;\n", s->name, s->initValue);
    }

    fprintf(fd, "        end\n");
    fprintf(fd, "        else \n");
    fprintf(fd, "        begin\n");

    cp = controlPointList;
    while (cp != NULL) {
        fprintf(fd, "            pbs_Y[%d] = ", cp->index);
        dumpVerilogExpression(fd, cp->f, "X");
        fprintf(fd, ";\n");
        cp = cp->next;
    }
    fprintf(fd, "\n");

    et = exThreadList;
    while (et != NULL) {
        fprintf(fd, "            pbs_active[%d] = ", et->index);
        dumpVerilogExpression(fd, et->active, "X");
        fprintf(fd, ";\n");
        et = et->next;
    }
    fprintf(fd, "\n");

    et = exThreadList;
    while (et != NULL) {
        fprintf(fd, "            pbs_okn[%d] = ", et->index);
        dumpVerilogExpression(fd, et->rec, "X");
        fprintf(fd, ";\n");
        et = et->next;
    }
    fprintf(fd, "\n");

    et = exThreadList;
    while (et != NULL) {
        fprintf(fd, "            pbs_reentry[%d] = pbs_active[%d] & pbs_okn[%d];\n", et->index, et->index, et->index);
        et = et->next;
    }
    fprintf(fd, "\n");

    et = exThreadList;
    while (et != NULL) {
        fprintf(fd, "            pbs_enable[%d] = ", et->index);
        dumpVerilogExpression(fd, et->enable, "X");
        fprintf(fd, ";\n");
        et = et->next;
    } 
    fprintf(fd, "\n");

    et = exThreadList;
    while (et != NULL) {
        fprintf(fd, "            pbs_ok[%d] = ", et->index);
        dumpVerilogExpression(fd, et->ok, "X");
        fprintf(fd, ";\n");
        et = et->next;
    }
    fprintf(fd, "\n");

    fprintf(fd, "            pbs_X = pbs_Y;\n\n"); 

    ap = actionPointList;
    while (ap != NULL) {
        fprintf(fd, "            pbs_A[%d] = ", ap->index);
        dumpVerilogExpression(fd, ap->f, "X");
        fprintf(fd, ";\n");
        ap = ap->next;
    }
    if (actionPointList != NULL) 
        fprintf(fd, "\n");

    ap = actionPointList;
    while (ap != NULL) {
        fprintf(fd, "            if (pbs_A[%d] == 1'b1)\n", ap->index);
        fprintf(fd, "            begin\n");
        dumpVerilogActionList(fd, ap->actionList);
        fprintf(fd, "            end\n\n");
        ap = ap->next;
    }
    if (actionPointList != NULL) 
        fprintf(fd, "\n");


    et = exThreadList;
    while (et != NULL) {
        fprintf(fd, "            ok = ok & (pbs_ok[%d] | ~pbs_enable[%d]) & ~pbs_reentry[%d];\n", et->index, et->index, et->index);
        et = et->next;
    }

    for (sm = stateMachineList; sm != NULL; sm = sm->next) {
        fprintf(fd, "            ok = ok & (|pbs_X[%d:%d]);\n",
                sm->cp[1], sm->cp[0]);
    }

    fprintf(fd, "\n");
    fprintf(fd, "        end\n");
    fprintf(fd, "    end\n\n");
}


void dumpVerilogEnd(FILE *fd) {
    fprintf(fd, "endmodule\n\n");
}


void dumpVerilogExpression(FILE *fd, treeNodePtr node, char *type) {

    controlPointPtr cp;
    wirePtr w;

    if (node == NULL) {
        fprintf(fd, "1'b0");
        return;
    }  

    if (isBoolExpr(node)) {
        dumpVerilogBoolExpression(fd, node);
        return;
    }

    switch (node->type) {
        case P_TN_CP:
            cp = (controlPointPtr) node->data;
            fprintf(fd, "pbs_%s[%d]", type, cp->index);
            break;
        case P_TN_AND_C:
            fprintf(fd, "(");
            dumpVerilogExpression(fd, node->left, type);
            fprintf(fd, " & ");
            dumpVerilogExpression(fd, node->right, type);
            fprintf(fd, ")");
            break;
        case P_TN_OR_C:
            fprintf(fd, "(");
            dumpVerilogExpression(fd, node->left, type);
            fprintf(fd, " | ");
            dumpVerilogExpression(fd, node->right, type);
            fprintf(fd, ")");
            break;
        case P_TN_NOT_C:
            fprintf(fd, "~(");
            dumpVerilogExpression(fd, node->left, type);
            fprintf(fd, ")");
            break;
        case P_TN_TEMP:
            dumpVerilogExpression(fd, node->left, type);
            break;
        case P_TN_TRUE_C:
            fprintf(fd, "1'b1");
            break;
        case P_TN_FALSE_C:
            fprintf(fd, "1'b0");
            break;
        case P_TN_WIRE_C:
            w = (wirePtr) node->data;
            fprintf(fd, "pbs_w[%d]", w->index); 
            break; 
        default:
            fatal_error("dumpVerilogExpression()"); 
    }
}

void dumpVerilogBoolExpression(FILE *fd, treeNodePtr node) {

    definePtr d;
    signalPtr s;

    switch (node->type) {
        case P_TN_DEFINE:
            d = (definePtr) node->data;
            fprintf(fd, "%s", d->name);
            break;
        case P_TN_SIGNAL:
        case P_TN_INTERNAL:
            s = (signalPtr) node->data;
            fprintf(fd, "%s", s->name);
            break;
        case P_TN_SIGNAL_ARRAY:
        case P_TN_INTERNAL_ARRAY:
            s = (signalPtr) node->data;
            fprintf(fd, "%s[", s->name);
            dumpVerilogIndex(fd, s->indexN);
            fprintf(fd, "]");
            break;
        case P_TN_EQUALEQUAL:
            fprintf(fd, "(");
            dumpVerilogBoolExpression(fd, node->left);
            fprintf(fd, " -~ ");
            dumpVerilogBoolExpression(fd, node->right);
            fprintf(fd, ")"); 
            break;
        case P_TN_NOTEQUAL:
            fprintf(fd, "(");
            dumpVerilogBoolExpression(fd, node->left);
            fprintf(fd, " ^ ");
            dumpVerilogBoolExpression(fd, node->right);
            fprintf(fd, ")");
            break; 
        case P_TN_AND_B:
            fprintf(fd, "(");
            dumpVerilogBoolExpression(fd, node->left);
            fprintf(fd, " & ");
            dumpVerilogBoolExpression(fd, node->right);
            fprintf(fd, ")");
            break;
        case P_TN_OR_B:
            fprintf(fd, "(");
            dumpVerilogBoolExpression(fd, node->left);
            fprintf(fd, " | ");
            dumpVerilogBoolExpression(fd, node->right);
            fprintf(fd, ")");
            break;
        case P_TN_NOT_B:
            fprintf(fd, "~");
            dumpVerilogBoolExpression(fd, node->left);
            break;
        case P_TN_CONSTANT:
            fprintf(fd, "%d", (int) node->data);
            break;
        default:
            fatal_error("dumpVerilogBoolExpression()"); 
    }
}

void dumpVerilogActionList(FILE *fd, actionPtr a) {

    actionPtr ap;

    for (ap = a; ap != NULL; ap = ap->next) {
        fprintf(fd, "                ");
        dumpVerilogActionExpression(fd, ap->lhs);
        fprintf(fd, " = ");
        dumpVerilogActionExpression(fd, ap->rhs);
        fprintf(fd, ";\n");
    } 
}

void dumpVerilogActionExpression(FILE *fd, actionNodePtr an) {

    switch (an->type) {
        case P_AN_PLUS:
            fprintf(fd, "(");
            dumpVerilogActionExpression(fd, an->left);
            fprintf(fd, " + ");
            dumpVerilogActionExpression(fd, an->right);
            fprintf(fd, ")");
            break;
        case P_AN_MINUS:
            fprintf(fd, "(");
            dumpVerilogActionExpression(fd, an->left);
            fprintf(fd, " - ");
            dumpVerilogActionExpression(fd, an->right);
            fprintf(fd, ")");
            break;
        case P_AN_CONSTANT:
            fprintf(fd, "%d", an->data.constant);
            break;
        case P_AN_SIGNAL:
            fprintf(fd, "%s", an->data.signal->name);
            break;
        case P_AN_SIGNAL_ARRAY:
            fprintf(fd, "%s[", an->data.signal->name);
            dumpVerilogIndex(fd, an->index);
            fprintf(fd, "]");
            break;
        default:
            fatal_error("dumpVerilogActionExpression()");
    }
}

void dumpVerilogIndex(FILE *fd, indexNodePtr in) {

    switch (in->type) {
        case P_IN_SIGNAL:
            fprintf(fd, "%s", in->data.signal->name);
            break;
        case P_IN_CONSTANT:
            fprintf(fd, "%d", in->data.constant);
            break;
        default:
            fatal_error("dumpVerilogIndex()");
    } 
}
